/*
 * Copyright (c) 1999 Apple Computer, Inc. All rights reserved.
 * Copyright (c) 1999 PADL Software Pty Ltd. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * "Portions Copyright (c) 1999 Apple Computer, Inc.  All Rights
 * Reserved.  This file contains Original Code and/or Modifications of
 * Original Code as defined in and that are subject to the Apple Public
 * Source License Version 1.0 (the 'License').  You may not use this file
 * except in compliance with the License.  Please obtain a copy of the
 * License at http://www.apple.com/publicsource and read it before using
 * this file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License."
 * 
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * C-library lookup routines which may be overridden by NetInfo.
 * Copyright (C) 1989 by NeXT, Inc.
 *
 */

/*
 * This module contains many of the get*ent routines in the C library.
 *
 * Search for:          to get:
 * GENERAL_STUFF        generally useful stuff 
 * PASSWD_STUFF         passwd specific stuff
 * GROUP_STUFF          group specific stuff
 * HOST_STUFF           hosts specific stuff
 * SERVICES_STUFF       services specific stuff
 * RPC_STUFF            rpc specific stuff
 * PROTOCOLS_STUFF      protocols specific stuff
 * NETWORKS_STUFF       networks specific stuff
 * PRINTER_STUFF        printer information
 * MOUNT_STUFF          mount information
 * BOOTP_STUFF          bootp information
 * BOOTPARAMS_STUFF     bootparams information
 * ALIAS_STUFF          alias information
 * NETGROUP_STUFF       netgroup information
 */

/*
 * GENERAL_STUFF
 */
#include <stdio.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <netdb.h>
#include <aliases.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <netinfo/ni.h>
#include <sys/socket.h>
#include <syslog.h>
#include <net/if.h>
#include <netinet/if_ether.h>
#include <ctype.h>
#include <nss.h>
#include <pthread.h>
#include "mm.h"

typedef struct ni_global_state_node {
    ni_id root_id;
    void *ni;
} ni_global_state_node;

typedef struct ni_global_state {
    ni_index len;
    ni_global_state_node *val;
} ni_global_state;

typedef struct ni_local_state_node {
    ni_id dir;
    ni_idlist children;
    ni_proplist_list entries;
    int havequicklist;		/* not sure if this is a good idea */
    int havequicklook;		/* this BS is only for 0.9 compatibility */
    ni_index which;
} ni_local_state_node;

typedef struct ni_local_state {
    ni_index which;
    ni_index len;
    ni_local_state_node *val;
} ni_local_state;

static pthread_mutex_t _nss_netinfo_mutex = PTHREAD_MUTEX_INITIALIZER;

#if 0
#define debug(str)   fprintf(stderr, "NSS_NETINFO %s\n", str);
#else
#define debug(str);
#endif

static int do_lock(
		      void
)
{
    return pthread_mutex_lock(&_nss_netinfo_mutex);
}

static int do_unlock(
			void
)
{
    return pthread_mutex_unlock(&_nss_netinfo_mutex);
}

#define MAX_INT_SIZE (3*sizeof(int) + 1)	/* portable, though liberal */

#define ID_NOBODY -2
#define MAXADDRLEN 32		/* Max ascii length of address */

#define UNINIT ((void *)(-1))
#define CANT_USE_QUICKPROPS (NSS_STATUS_RETURN + 1)
#define CHECK_RC(qp, pl, rc, f) do { \
		rc = (f); \
		if (rc != NSS_STATUS_SUCCESS) { \
			if (qp == NULL) { \
				ni_proplist_free(&pl); \
			} \
			return rc; \
		} \
	} while (0)

static const ni_name NAME_NAME = "name";
static const ni_name NAME_USERS = "users";
static const ni_name NAME_GID = "gid";

static const char STR_NULL[] = "";
static const char *STRLST_NULL[] =
{NULL};

static int fully_expanded = 0;	/* All parents in nichain */

static struct ni_global_state nichain =
{NI_INDEX_NULL, UNINIT};

/*
 * The idea behind the quicklist was to get all of the information in
 * a single RPC call. While this may be easier on the network, it is
 * not very good for the server because it ties it up. We are turning
 * this feature off here by not defining USEQUICKLIST. It may disappear
 * from the server too.
 */
#define NENTRIES(state)       ((state).havequicklist ? \
			       (state).entries.nipll_len : \
			       (state).children.niil_len)

static ni_name commafy(ni_namelist nl);

#ifdef notdef
static struct mntent *_new_ni_getmntent(FILE *);
static FILE *_new_ni_setmntent(char *, char *);
static int _new_ni_endmntent(FILE * f);
static struct mntent *_old_ni_getmntent(FILE *);
static FILE *_old_ni_setmntent(char *, char *);
static int _old_ni_endmntent(FILE * f);

extern const char *ni_error_msg(ni_status);
#endif

static char *
 itoa(
	 char *str,
	 int num
)
{
    char buf[MAX_INT_SIZE + 1];
    char *p = buf;
    int sign = 0;
    char *s;

    if (num < 0) {
	sign++;
	num = -num;
    }
    while (num > 0) {
	*p++ = '0' + (num % 10);
	num /= 10;
    }
    if (p == buf) {
	*p++ = '0';
    }
    /* reverse */
    s = str;
    if (sign) {
	*s++ = '-';
    }
    for (p--; p >= buf; p--) {
	*s++ = *p;
    }
    *s = 0;
    return (str);
}

static int _ni_running(
			  void
)
{
    void *local_ni;
    ni_id root_id;
    ni_index i;

    if (nichain.val != UNINIT) {
	return (nichain.val != NULL);
    }
    local_ni = ni_new(NULL, ".");
    if (local_ni == NULL) {
	return (0);
    }
    if (ni_root(local_ni, &root_id) != NI_OK) {
	ni_free(local_ni);
	nichain.val = NULL;
	return (0);
    }
    if (nichain.val != UNINIT) {
	for (i = 0; i < nichain.len; i++) {
	    ni_free(nichain.val[i].ni);
	}
	MM_FREE(nichain.val);
    }
    MM_ALLOC(nichain.val);
    nichain.val[0].ni = local_ni;
    nichain.val[0].root_id = root_id;
    nichain.len = 1;
    return (1);
}

/* Majka 2/22/94 - log a list of server addresses/tags */
static void logchain(void)
{
    int i;
    ni_status status;
    struct sockaddr_in sin;
    ni_name tag;

    if (!_ni_running()) {
	syslog(LOG_INFO, "NetInfo is not running");
	return (void) NULL;
    }
    for (i = 0; i < nichain.len; i++) {
	status = ni_addrtag(nichain.val[i].ni, &sin, &tag);
	if (status != NI_OK) {
	    syslog(LOG_INFO,
	    "%d: Can't get address and tag: %s", i + 1, ni_error(status));
	} else {
	    syslog(LOG_INFO,
		   "%d: address %s tag: %s", i + 1,
		   inet_ntoa(sin.sin_addr), tag);
	}
    }

    return (void) NULL;
}

static unsigned getdirinstance(
				  void *ni,
				  ni_id * root_id,
				  ni_name dir
)
{
    ni_id nid;
    ni_idlist bork;

    /*
     * Return zero in case of failure: this will be consistent for 
     * domains that don't have the given directory.
     */

    if ((ni_lookup(ni, root_id, NAME_NAME, dir, &bork) != NI_OK)
	|| !bork.niil_len) {
	return 0;
    }
    nid.nii_object = bork.niil_val[0];
    ni_idlist_free(&bork);
    if (ni_self(ni, &nid) != NI_OK) {
	return random();
    }
    return (nid.nii_instance);
}

static ni_index
 getchecksum(
		void *ni
)
{
    ni_status status;
    ni_proplist pl;
    ni_index sum;
    ni_namelist nl;
    ni_index i;
    static const ni_name NAME_CHECKSUM = "checksum";

    /*
     * Set to random for case of failure: random checksums will
     * almost never match one another.
     */
    sum = random();
    status = ni_statistics(ni, &pl);

    if (status != NI_OK) {
	return (sum);
    }
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_CHECKSUM)) {
	    nl = pl.nipl_val[i].nip_val;
	    if (nl.ninl_len > 0) {
		sum = atoi(nl.ninl_val[0]);
	    }
	    break;
	}
    }
    ni_proplist_free(&pl);
    return (sum);
}

/* GRS 2/26/92 - Added dir arg; if non-null, name of dir to get dirID for */

static void ni_getchecksums(
			       unsigned *lenp,
			       unsigned **sumsp,
			       ni_name dir
)
{
    unsigned len;
    unsigned *sums;
    void *ni;
    ni_id root;

    len = 0;
    sums = NULL;
    if (!_ni_running()) {
	*lenp = len;
	*sumsp = sums;
	return;
    }
    while (len < nichain.len) {
	MM_GROW_ARRAY(sums, len);
	if (dir) {
	    sums[len] = getdirinstance(nichain.val[len].ni,
				       &nichain.val[len].root_id, dir);
	} else {
	    sums[len] = getchecksum(nichain.val[len].ni);
	}
	len++;

	if ((len + 1 < nichain.len) || fully_expanded) {
	    continue;
	}
	/*
	 * Need to grow it
	 */
	ni = ni_new(nichain.val[nichain.len - 1].ni, "..");
	if (ni == NULL) {
	    continue;
	}
	if (ni_root(ni, &root) == NI_OK) {
	    MM_GROW_ARRAY(nichain.val, nichain.len);
	    nichain.val[nichain.len].ni = ni;
	    nichain.val[nichain.len].root_id = root;
	    nichain.len++;
	} else {
	    ni_free(ni);
	}
    }
    *lenp = len;
    *sumsp = sums;
}

static enum nss_status name_get(
				   ni_namelist * nl,
				   char **val,
				   char **buffer,
				   size_t * buflen,
				   int *errnop
)
{
    int vallen;

    if (nl->ninl_len == 0) {
	*val = NULL;
	return NSS_STATUS_SUCCESS;
    }
    vallen = strlen(nl->ninl_val[0]);
    if (*buflen < (size_t) (vallen + 1)) {
	*errnop = ERANGE;
	return NSS_STATUS_TRYAGAIN;
    }
    *val = *buffer;
    strncpy(*val, nl->ninl_val[0], vallen);
    (*val)[vallen] = '\0';

    *buffer += vallen + 1;
    *buflen -= vallen + 1;

    return NSS_STATUS_SUCCESS;
}

static enum nss_status int_get(
				  ni_namelist * nl,
				  int *ival
)
{
    int val;
    int sign;
    ni_name nm;

    val = 0;
    sign = 1;
    if (nl->ninl_len > 0) {
	for (nm = nl->ninl_val[0]; *nm; nm++) {
	    if (val == 0 && *nm == '-') {
		sign = -sign;
	    } else if (isdigit(*nm)) {
		val *= 10;
		val += *nm - '0';
	    } else {
		return;
	    }
	}
    }
    *ival = sign * val;
    return NSS_STATUS_SUCCESS;
}

#define alignof(x)    __alignof__(x)

#define align(ptr, blen)              do { \
		char *qtr = ptr; \
		ptr += (alignof(char *)) - 1; \
		ptr -= ((ptr - (char *)NULL) % alignof(char *)); \
		blen -= (ptr - qtr); \
            } while (0)

#define bytesleft(ptr, blen)    (blen - alignof(char *) + 1)

static enum nss_status namelist_get(
				       ni_namelist * nl,
				       char ***valptr,
				       char **pbuffer,
				       size_t * pbuflen,
				       int *errnop
)
{
    ni_index i;
    char **p = NULL;
    register int buflen = *pbuflen;
    register char *buffer = *pbuffer;

    if (bytesleft(buffer, buflen) < (nl->ninl_len + 1) * sizeof(char *)) {
	*errnop = ERANGE;
	return NSS_STATUS_TRYAGAIN;
    }
    align(buffer, buflen);

    p = *valptr = (char **) buffer;

    buffer += (nl->ninl_len + 1) * sizeof(char *);
    buflen -= (nl->ninl_len + 1) * sizeof(char *);

    if (nl->ninl_len == 0) {
	*p = NULL;
	*pbuffer = buffer;
	*pbuflen = buflen;
	return NSS_STATUS_SUCCESS;
    }
    for (i = 0; i < nl->ninl_len; i++) {
	int vallen;
	char *elt = NULL;

	vallen = strlen(nl->ninl_val[i]);
	if (buflen < (size_t) (vallen + 1)) {
	    *errnop = ERANGE;
	    return NSS_STATUS_TRYAGAIN;
	}
	elt = buffer;
	buffer += vallen + 1;
	buflen -= vallen + 1;

	strncpy(elt, nl->ninl_val[i], vallen);
	elt[vallen] = '\0';
	*p = elt;
	p++;
    }

    *p = NULL;
    *pbuffer = buffer;
    *pbuflen = buflen;

    return NSS_STATUS_SUCCESS;
}

static void freelist(
			char **l
)
{
    int i;

    for (i = 0; l[i] != NULL; i++) {
	free(l[i]);
    }
    free(l);
}

static enum nss_status addr_get(
				   ni_namelist * nl,
				   char ***valptr,
				   char **pbuffer,
				   size_t * pbuflen,
				   int *errnop
)
{
    u_long addr;
    ni_index i;
    char **p = NULL;
    register int buflen = *pbuflen;
    register char *buffer = *pbuffer;

    if (buflen < (nl->ninl_len + 1) * sizeof(char *)) {
	*errnop = ERANGE;
	return NSS_STATUS_TRYAGAIN;
    }
    align(buffer, buflen);

    p = *valptr = (char **) buffer;

    buffer += (nl->ninl_len + 1) * sizeof(char *);
    buflen -= (nl->ninl_len + 1) * sizeof(char *);

    if (nl->ninl_len == 0) {
	*p = NULL;
	*pbuffer = buffer;
	*pbuflen = buflen;
	return NSS_STATUS_SUCCESS;
    }
    for (i = 0; i < nl->ninl_len; i++) {
	char *elt = NULL;

	if (buflen < sizeof(u_long)) {
	    *errnop = ERANGE;
	    return NSS_STATUS_TRYAGAIN;
	}
	elt = buffer;
	buffer += sizeof(u_long);
	buflen -= sizeof(u_long);

	addr = inet_addr(nl->ninl_val[i]);
	memcpy(elt, &addr, sizeof(addr));
	*p = elt;
	p++;
    }

    *p = NULL;
    *pbuffer = buffer;
    *pbuflen = buflen;

    return NSS_STATUS_SUCCESS;
}

static void sorry(
		     char *func
)
{
    (void) fprintf(stderr, "sorry, %s not supported with NetInfo (yet)\n",
		   func);
}

static enum nss_status do_init(
				  const ni_name name,
				  ni_local_state * state
)
{
    ni_idlist match;

    if (!_ni_running()) {
	return (NSS_STATUS_UNAVAIL);
    }
    if (state->val != NULL) {
	return (NSS_STATUS_SUCCESS);
    }
    if (ni_lookup(nichain.val[0].ni, &nichain.val[0].root_id,
		  NAME_NAME, name, &match) != NI_OK) {
	return (NSS_STATUS_UNAVAIL);
    }
    state->len = 1;
    state->which = 0;
    MM_ALLOC(state->val);
    state->val[0].dir.nii_object = match.niil_val[0];
    state->val[0].which = NI_INDEX_NULL;
    state->val[0].havequicklook = 1;
    ni_idlist_free(&match);
    return (NSS_STATUS_SUCCESS);
}

static enum nss_status do_end(
				 ni_local_state * state
)
{
    return NSS_STATUS_SUCCESS;
}

static enum nss_status do_set(
				 ni_local_state * state
)
{
    ni_index i;

    for (i = 0; i < state->len; i++) {
	if (state->val[i].which != NI_INDEX_NULL) {
	    ni_idlist_free(&state->val[i].children);
	    if (state->val[i].havequicklist) {
		ni_proplist_list_free(&state->val[i].entries);
	    }
	    state->val[i].which = NI_INDEX_NULL;
#ifdef USEQUICKLIST
	    state->val[i].havequicklist = 1;
#else
	    state->val[i].havequicklist = 0;
#endif
	}
    }
    state->which = 0;
    return NSS_STATUS_SUCCESS;
}

static enum nss_status do_more(
				  const ni_name_const name,
				  ni_local_state * state,
				  ni_index which
)
{
    void *ni;
    ni_idlist match;
    ni_id root;

    if (nichain.len <= which + 1) {
	/*
	 * Need to grow it
	 */
	ni = ni_new(nichain.val[nichain.len - 1].ni, "..");
	if (ni == NULL) {
	    return (NSS_STATUS_UNAVAIL);
	}
	if (ni_root(ni, &root) != NI_OK) {
	    ni_free(ni);
	    return (NSS_STATUS_UNAVAIL);
	}
	MM_GROW_ARRAY(nichain.val, nichain.len);
	nichain.val[nichain.len].ni = ni;
	nichain.val[nichain.len].root_id = root;
	nichain.len++;
    } else {
	ni = nichain.val[which + 1].ni;
	root = nichain.val[which + 1].root_id;
    }
    if (state && which + 1 >= state->len) {
	if (ni_lookup(ni, &root, NAME_NAME, name, &match) != NI_OK) {
	    return (NSS_STATUS_UNAVAIL);
	}
	MM_GROW_ARRAY(state->val, state->len);
	state->val[state->len].dir.nii_object = match.niil_val[0];
	state->val[state->len].which = NI_INDEX_NULL;
	state->val[state->len].havequicklook = 1;
	ni_idlist_free(&match);
	state->len++;
    }
    if (state && state->val[which + 1].which != NI_INDEX_NULL) {
	state->val[which + 1].which = NI_INDEX_NULL;
	ni_idlist_free(&state->val[which + 1].children);
	if (state->val[which + 1].havequicklist) {
	    ni_proplist_list_free(&state->val[which + 1].entries);
	}
    }
    return (NSS_STATUS_SUCCESS);
}

static enum nss_status do_get(
				 const ni_name_const name,
				 ni_local_state * state,
				 const ni_name_const prop_name,
				 const ni_name_const prop_val,
				 enum nss_status (*readit) (),
				 void *arg,
				 void *result,
				 char *buffer,
				 size_t buflen,
				 int *errnop
)
{
    ni_idlist match;
    enum nss_status res = CANT_USE_QUICKPROPS;
    ni_id id;
    ni_status status = NI_FAILED;
    ni_index i;
    void *ni;
    ni_proplist props;

    if (state->val[0].which == NI_INDEX_NULL) {
	do_set(state);
    }
    for (i = 0; i < state->len; i++) {
	ni = nichain.val[i].ni;
	id = state->val[i].dir;
	if (state->val[i].havequicklook) {
	    status = ni_lookupread(ni, &id, prop_name,
				   prop_val, &props);
	    if (status == NI_FAILED) {
		state->val[i].havequicklook = 0;
	    }
	}
	do {
	    if (!state->val[i].havequicklook) {
		status = ni_lookup(ni, &id, prop_name,
				   prop_val, &match);
	    }
	    if (status != NI_OK && status != NI_NODIR) {
		return (NSS_STATUS_UNAVAIL);
	    }
	    if (status == NI_OK) {
		if (state->val[i].havequicklook) {
		    res = (*readit) (ni, NULL, arg,
				 &props, result, buffer, buflen, errnop);
		    ni_proplist_free(&props);
		} else {
		    res = (*readit) (ni, &match, arg,
				   NULL, result, buffer, buflen, errnop);
		    ni_idlist_free(&match);
		}
		if (res == NSS_STATUS_SUCCESS) {
		    return (res);
		}
		state->val[i].havequicklook = 0;
	    }
	} while (res == CANT_USE_QUICKPROPS && status == NI_OK);
	if ((res = do_more(name, state, i)) != NSS_STATUS_SUCCESS) {
	    return res;
	}
    }
    return (NSS_STATUS_UNAVAIL);
}

static enum nss_status do_getent(
				    ni_name name,
				    ni_local_state * state,
				    enum nss_status (*readit) (),
				    void *arg,
				    void *result,
				    char *buffer,
				    size_t buflen,
				    int *errnop
)
{
    ni_idlist match;
    ni_local_state_node *lsn;
    enum nss_status rc;

    if (state->val[0].which == NI_INDEX_NULL) {
	do_set(state);
#ifdef USEQUICKLIST
	if (ni_listall(nichain.val[state->which].ni,
		       &state->val[state->which].dir,
		       &state->val[state->which].entries) == NI_OK) {
	    state->val[state->which].havequicklist = 1;
	} else
#endif
	{
	    state->val[state->which].havequicklist = 0;
	}
	if (ni_children(nichain.val[state->which].ni,
			&state->val[state->which].dir,
			&state->val[state->which].children) == NI_OK) {
	    state->val[state->which].which = 0;
	}
    }
    while (NENTRIES(state->val[state->which]) <=
	   state->val[state->which].which) {
	if ((rc = do_more(name, state, state->which)) != NSS_STATUS_SUCCESS) {
	    return (rc);
	}
	state->which++;
#ifdef USEQUICKLIST
	if (ni_listall(nichain.val[state->which].ni,
		       &state->val[state->which].dir,
		       &state->val[state->which].entries) == NI_OK) {
	    state->val[state->which].havequicklist = 1;
	} else
#endif
	{
	    state->val[state->which].havequicklist = 0;
	}
	if (ni_children(nichain.val[state->which].ni,
			&state->val[state->which].dir,
			&state->val[state->which].children)
	    != NI_OK) {
	    state->val[state->which].children.niil_len = 0;
	    state->val[state->which].children.niil_val = NULL;
	}
	state->val[state->which].which = 0;
    }
    lsn = &state->val[state->which];
    match.niil_val = &lsn->children.niil_val[lsn->which++];
    match.niil_len = 1;
    return ((*readit) (nichain.val[state->which].ni, &match, arg,
		       (lsn->havequicklist ?
			&lsn->entries.nipll_val[lsn->which - 1] : NULL),
		       result, buffer, buflen, errnop));
}

/*
 * PASSWD_STUFF
 */
static const ni_name NAME_PASSWD = "passwd";
static const ni_name NAME_UID = "uid";
static const ni_name NAME_REALNAME = "realname";
static const ni_name NAME_HOME = "home";
static const ni_name NAME_SHELL = "shell";

static enum nss_status pw_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);
#define pw_init() do_init(NAME_USERS, &pw_state)

static ni_local_state pw_state =
{NI_INDEX_NULL};
static struct passwd pw =
{
    (char *) STR_NULL, (char *) STR_NULL, ID_NOBODY, ID_NOBODY,
#ifndef BSD44PASSWD
    (char *) STR_NULL, (char *) STR_NULL, (char *) STR_NULL
#else
    0, (char *) STR_NULL, (char *) STR_NULL, (char *) STR_NULL,
    (char *) STR_NULL
#endif
};

enum nss_status _nss_netinfo_getpwuid_r(
					   uid_t uid,
					   struct passwd *result,
					   char *buffer,
					   size_t buflen,
					   int *errnop
)
{
    char uidbuf[MAX_INT_SIZE + 1];
    ni_name uidnm;
    enum nss_status rc;

    do_lock();
    rc = pw_init();
    if (rc == NSS_STATUS_SUCCESS) {
	uidnm = itoa(uidbuf, uid);
	rc = do_get(NAME_USERS, &pw_state, NAME_UID, uidnm, pw_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_getpwnam_r(
					   const char *name,
					   struct passwd *result,
					   char *buffer,
					   size_t buflen,
					   int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = pw_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_get(NAME_USERS, &pw_state, NAME_NAME, name, pw_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_endpwent(
					 void
)
{
    enum nss_status rc;

    do_lock();
    rc = pw_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&pw_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setpwent(
					 void
)
{
    enum nss_status rc;

    do_lock();
    rc = pw_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&pw_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getpwent_r(
					   struct passwd *result,
					   char *buffer,
					   size_t buflen,
					   int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = pw_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_USERS, &pw_state, pw_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}

static enum nss_status pw_read(
				  void *ni,
				  ni_idlist * match,
				  void *unused,
				  ni_proplist * quickprops,
				  void *result,
				  char *buffer,
				  size_t buflen,
				  int *errnop
)
{
    ni_proplist pl;
    ni_index i;
    ni_id id;
    struct passwd *pw = (struct passwd *) result;
    enum nss_status rc;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    return (NSS_STATUS_TRYAGAIN);
	}
    } else {
	pl = *quickprops;
    }
    memset(pw, 0, sizeof *pw);
    pw->pw_gecos = (char *) STR_NULL;
    pw->pw_passwd = (char *) STR_NULL;
    pw->pw_name = (char *) STR_NULL;
    pw->pw_shell = (char *) STR_NULL;
    pw->pw_dir = (char *) STR_NULL;
    pw->pw_uid = ID_NOBODY;
    pw->pw_gid = ID_NOBODY;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_REALNAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &pw->pw_gecos, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_PASSWD)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &pw->pw_passwd, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_NAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &pw->pw_name, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_SHELL)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &pw->pw_shell, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_HOME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &pw->pw_dir, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_UID)) {
	    int_get(&pl.nipl_val[i].nip_val, &pw->pw_uid);
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_GID)) {
	    int_get(&pl.nipl_val[i].nip_val, &pw->pw_gid);
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return (NSS_STATUS_SUCCESS);
}

/*
 * GROUP_STUFF
 */
static const ni_name NAME_GROUPS = "groups";

static enum nss_status gr_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);
#define gr_init() do_init(NAME_GROUPS, &gr_state)

static ni_local_state gr_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_getgrgid_r(
					   gid_t gid,
					   struct group *result,
					   char *buffer,
					   size_t buflen,
					   int *errnop
)
{
    char gidbuf[MAX_INT_SIZE + 1];
    ni_name gidnm;
    enum nss_status rc;

    do_lock();
    rc = gr_init();
    if (rc == NSS_STATUS_SUCCESS) {
	gidnm = itoa(gidbuf, gid);
	rc = do_get(NAME_GROUPS, &gr_state, NAME_GID,
		    gidnm, gr_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_getgrnam_r(
					   const char *name,
					   struct group *result,
					   char *buffer,
					   size_t buflen,
					   int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = gr_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_get(NAME_GROUPS, &gr_state, NAME_NAME,
		    name, gr_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_endgrent(
					 void
)
{
    enum nss_status rc;

    do_lock();
    rc = gr_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&gr_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setgrent(
					 void
)
{
    enum nss_status rc;

    do_lock();
    rc = gr_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&gr_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getgrent_r(
					   struct group *result,
					   char *buffer,
					   size_t buflen,
					   int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = gr_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_GROUPS, &gr_state, gr_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}

static enum nss_status gr_read(
				  void *ni,
				  ni_idlist * match,
				  void *ununsed,
				  ni_proplist * quickprops,
				  void *result,
				  char *buffer,
				  size_t buflen,
				  int *errnop
)
{
    ni_proplist pl;
    ni_index i;
    ni_id id;
    struct group *gr = (struct group *) result;
    enum nss_status rc;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    return (NSS_STATUS_TRYAGAIN);
	}
    } else {
	pl = *quickprops;
    }
    gr->gr_name = (char *) STR_NULL;
    gr->gr_passwd = (char *) STR_NULL;
    gr->gr_mem = (char **) STRLST_NULL;
    gr->gr_gid = ID_NOBODY;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &gr->gr_name, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_PASSWD)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &gr->gr_passwd, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_USERS)) {
	    CHECK_RC(quickprops, pl, rc, namelist_get(&pl.nipl_val[i].nip_val, &gr->gr_mem, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_GID)) {
	    int_get(&pl.nipl_val[i].nip_val, &gr->gr_gid);
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return NSS_STATUS_SUCCESS;
}

/*
 * HOST_STUFF
 */
static const ni_name NAME_MACHINES = "machines";
static const ni_name NAME_IP_ADDRESS = "ip_address";

/*extern int h_errno; */
static enum nss_status h_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);

#define h_init() do_init(NAME_MACHINES, &h_state)

static ni_local_state h_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_gethostbyname_r(
						char *name,
						struct hostent *result,
						char *buffer,
						size_t buflen,
						int *errnop,
						int *h_errnop
)
{
    ni_name hnm;
    enum nss_status rc;

    do_lock();
    rc = h_init();
    if (rc != NSS_STATUS_SUCCESS) {
	*h_errnop = NO_RECOVERY;
	do_unlock();
	return rc;
    }
    hnm = name;
    *h_errnop = HOST_NOT_FOUND;
    rc = do_get(NAME_MACHINES, &h_state, NAME_NAME, hnm, h_read,
		h_errnop, result, buffer, buflen, errnop);
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_gethostbyaddr_r(
						char *addr,
						int len,
						int type,
						struct hostent *result,
						char *buffer,
						size_t buflen,
						int *errnop,
						int *h_errnop
)
{
    ni_name nm;
    struct in_addr in;
    enum nss_status rc;

    do_lock();
    rc = h_init();
    if (rc != NSS_STATUS_SUCCESS) {
	*h_errnop = NO_RECOVERY;
	do_unlock();
	return rc;
    }
    if (len != sizeof(u_long) || type != AF_INET) {
	*h_errnop = NO_RECOVERY;
	do_unlock();
	return NSS_STATUS_UNAVAIL;
    }
    memcpy(&in.s_addr, addr, sizeof(in.s_addr));
    nm = inet_ntoa(in);
    *h_errnop = HOST_NOT_FOUND;
    rc = do_get(NAME_MACHINES, &h_state, NAME_IP_ADDRESS, nm, h_read,
		h_errnop, result, buffer, buflen, errnop);
    do_unlock();
    return rc;
}
enum nss_status _nss_netinfo_endhostent(
					   void
)
{
    enum nss_status rc;

    do_lock();
    rc = h_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&h_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_sethostent(
					   void
)
{
    enum nss_status rc;

    do_lock();
    rc = h_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&h_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_gethostent_r(
					     struct hostent *result,
					     char *buffer,
					     size_t buflen,
					     int *errnop,
					     int *h_errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = h_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_MACHINES, &h_state, h_read, NULL, result, buffer, buflen, errnop);
    } else {
	*h_errnop = NO_RECOVERY;
    }
    do_unlock();
    return rc;
}

static enum nss_status h_read(
				 void *ni,
				 ni_idlist * match,
				 void *unused,
				 ni_proplist * quickprops,
				 void *result,
				 char *buffer,
				 size_t buflen,
				 int *errnop
)
{
    struct hostent *h = (struct hostent *) result;
    int *h_errnop = (int *) unused;
    enum nss_status rc;
    ni_proplist pl;
    ni_index i;
    ni_namelist nl;
    ni_id id;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    *h_errnop = TRY_AGAIN;
	    return NSS_STATUS_TRYAGAIN;
	}
    } else {
	pl = *quickprops;
    }
    h->h_name = (char *) STR_NULL;
    h->h_aliases = (char **) STRLST_NULL;
    h->h_addr_list = (char **) STRLST_NULL;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &h->h_name, &buffer, &buflen, errnop));
	    nl = pl.nipl_val[i].nip_val;
	    if (nl.ninl_len > 0) {
		nl.ninl_len--;
		nl.ninl_val++;
		CHECK_RC(quickprops, pl, rc, namelist_get(&nl, &h->h_aliases, &buffer, &buflen, errnop));
	    }
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_IP_ADDRESS)) {
	    CHECK_RC(quickprops, pl, rc, addr_get(&pl.nipl_val[i].nip_val, &h->h_addr_list, &buffer, &buflen, errnop));
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    *h_errnop = NETDB_SUCCESS;
    return NSS_STATUS_SUCCESS;
}

/*
 * SERVICE_STUFF
 */
static const ni_name NAME_SERVICES = "services";
static const ni_name NAME_PROTOCOL = "protocol";
static const ni_name NAME_PORT = "port";

static enum nss_status s_read(void *, ni_idlist *, char *, ni_proplist *, void *, char *, size_t, int *);
#define s_init() do_init(NAME_SERVICES, &s_state)

static ni_local_state s_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_getservbyname_r(
						char *name,
						char *proto,
						struct servent *result,
						char *buffer,
						size_t buflen,
						int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = s_init();
    if (rc == NSS_STATUS_SUCCESS) {
	if (!strlen(proto)) {
	    proto = NULL;	/* User supplied NULL, encoded as "" */
	}
	rc = do_get(NAME_SERVICES, &s_state, NAME_NAME,
		   name, s_read, proto, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_getservbyport_r(
						int port,
						char *proto,
						struct servent *result,
						char *buffer,
						size_t buflen,
						int *errnop
)
{
    ni_name nm;
    char buf[MAXADDRLEN + 1];
    enum nss_status rc;

    do_lock();
    rc = s_init();
    if (rc == NSS_STATUS_SUCCESS) {
	if (!strlen(proto)) {
	    proto = NULL;	/* User supplied NULL, encoded as "" */
	}
	nm = itoa(buf, port);
	rc = do_get(NAME_SERVICES, &s_state, NAME_PORT,
		   nm, s_read, proto, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getservent_r(
					     struct servent *result,
					     char *buffer,
					     size_t buflen,
					     int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = s_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = s_read(NULL, NULL, NULL, NULL, result, buffer, buflen, errnop);
	if (rc != NSS_STATUS_SUCCESS) {
	    rc = do_getent(NAME_SERVICES, &s_state, s_read, NULL, result, buffer, buflen, errnop);
	}
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_endservent(
					   void
)
{
    enum nss_status rc;

    do_lock();
    rc = s_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&s_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setservent(
					   void
)
{
    enum nss_status rc;

    do_lock();
    rc = s_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&s_state);
    }
    do_unlock();
    return rc;
}

static enum nss_status s_read(
				 void *ni,
				 ni_idlist * il,
				 char *proto,
				 ni_proplist * quickprops,
				 void *result,
				 char *buffer,
				 size_t buflen,
				 int *errnop
)
{
    struct servent *s = (struct servent *) result;
    enum nss_status rc;
    ni_proplist pl;
    ni_index i;
    ni_index j;
    ni_index k;
    ni_id id;
    ni_namelist nl;
    /*XXX */ static ni_namelist protolist;

    if (ni == NULL) {
	if (protolist.ninl_len == 0) {
	    return (NSS_STATUS_NOTFOUND);
	}
	/*XXX fill in rest of servent */
	CHECK_RC(quickprops, pl, rc, name_get(&protolist, &s->s_proto, &buffer, &buflen, errnop));
	ni_namelist_delete(&protolist, 0);
	return NSS_STATUS_SUCCESS;
    }
    if (il == NULL) {
	return (CANT_USE_QUICKPROPS);
    }
    ni_namelist_free(&protolist);
    for (j = 0; j < il->niil_len; j++) {
	id.nii_object = il->niil_val[j];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    continue;
	}
	s->s_name = (char *) STR_NULL;
	s->s_aliases = (char **) STRLST_NULL;
	s->s_proto = (char *) STR_NULL;
	for (i = 0; i < pl.nipl_len; i++) {
	    if (ni_name_match(pl.nipl_val[i].nip_name,
			      NAME_NAME)) {
		CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &s->s_name, &buffer, &buflen, errnop));
		nl = pl.nipl_val[i].nip_val;
		if (nl.ninl_len > 0) {
		    nl.ninl_len--;
		    nl.ninl_val++;
		    CHECK_RC(quickprops, pl, rc, namelist_get(&nl, &s->s_aliases, &buffer, &buflen, errnop));
		}
	    } else if (ni_name_match(pl.nipl_val[i].nip_name,
				     NAME_PORT)) {
		int_get(&pl.nipl_val[i].nip_val, &s->s_port);
		s->s_port = htons(s->s_port);
	    } else if (ni_name_match(pl.nipl_val[i].nip_name,
				     NAME_PROTOCOL)) {
		nl = pl.nipl_val[i].nip_val;
		if (proto == NULL) {
		    protolist = ni_namelist_dup(nl);
		    CHECK_RC(quickprops, pl, rc, name_get(&protolist, &s->s_proto, &buffer, &buflen, errnop));
		    ni_namelist_delete(&protolist, 0);
		} else {
		    memset(&protolist, 0, sizeof(protolist));
		    for (k = 0; k < nl.ninl_len; k++) {
			if (ni_name_match(nl.ninl_val[k],
					  proto)) {
			    (s->s_proto =
			     ni_name_dup(proto));
			    break;
			}
		    }
		}
	    }
	}
	ni_proplist_free(&pl);
	if (s->s_proto != (char *) STR_NULL) {
	    return (NSS_STATUS_SUCCESS);
	}
    }
    return (NSS_STATUS_NOTFOUND);
}

/*
 * RPC_STUFF
 */
static const ni_name NAME_RPCS = "rpcs";
static const ni_name NAME_NUMBER = "number";

static enum nss_status r_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);
#define r_init() do_init(NAME_RPCS, &r_state)

static ni_local_state r_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_getrpcbynumber_r(
						 int num,
						 struct rpcent *result,
						 char *buffer,
						 size_t buflen,
						 int *errnop
)
{
    char buf[MAX_INT_SIZE + 1];
    ni_name nm;
    enum nss_status rc;

    do_lock();
    rc = r_init();
    if (rc == NSS_STATUS_SUCCESS) {
	nm = itoa(buf, num);
	rc = do_get(NAME_RPCS, &r_state, NAME_NUMBER,
		    nm, r_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_getrpcbyname_r(
					       const char *name,
					       struct rpcent *result,
					       char *buffer,
					       size_t buflen,
					       int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = r_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_get(NAME_RPCS, &r_state, NAME_NAME,
		    name, r_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_endrpcent(
					  void
)
{
    enum nss_status rc;

    do_lock();
    rc = r_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&r_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setrpcent(
					  void
)
{
    enum nss_status rc;

    do_lock();
    rc = r_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&r_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getrpcent_r(
					    struct group *result,
					    char *buffer,
					    size_t buflen,
					    int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = r_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_RPCS, &r_state, r_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}


static enum nss_status r_read(
				 void *ni,
				 ni_idlist * match,
				 void *unused,
				 ni_proplist * quickprops,
				 void *result,
				 char *buffer,
				 size_t buflen,
				 int *errnop
)
{
    struct rpcent *r = (struct rpcent *) result;
    enum nss_status rc;
    ni_proplist pl;
    ni_index i;
    ni_namelist nl;
    ni_id id;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    return (NSS_STATUS_TRYAGAIN);
	}
    } else {
	pl = *quickprops;
    }
    r->r_name = (char *) STR_NULL;
    r->r_aliases = (char **) STRLST_NULL;
    r->r_number = 0;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &r->r_name, &buffer, &buflen, errnop));
	    nl = pl.nipl_val[i].nip_val;
	    if (nl.ninl_len > 0) {
		nl.ninl_len--;
		nl.ninl_val++;
		CHECK_RC(quickprops, pl, rc, namelist_get(&nl, &r->r_aliases, &buffer, &buflen, errnop));
	    }
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_NUMBER)) {
	    int_get(&pl.nipl_val[i].nip_val, &r->r_number);
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return (NSS_STATUS_SUCCESS);
}

/*
 * PROTOCOLS_STUFF
 */
static const ni_name NAME_PROTOCOLS = "protocols";

static enum nss_status p_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);
#define p_init() do_init(NAME_PROTOCOLS, &p_state)
static ni_local_state p_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_getprotobynumber_r(
						   u_long num,
						 struct protoent *result,
						   char *buffer,
						   size_t buflen,
						   int *errnop
)
{
    char buf[MAX_INT_SIZE + 1];
    ni_name nm;
    enum nss_status rc;

    do_lock();
    rc = p_init();
    if (rc == NSS_STATUS_SUCCESS) {
	nm = itoa(buf, num);
	rc = do_get(NAME_PROTOCOLS, &p_state, NAME_NUMBER,
		    nm, p_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_getprotobyname_r(
						 const char *name,
						 struct protoent *result,
						 char *buffer,
						 size_t buflen,
						 int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = p_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_get(NAME_PROTOCOLS, &p_state, NAME_NAME,
		    name, p_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_endprotoent(
					    void
)
{
    enum nss_status rc;

    do_lock();
    rc = p_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&p_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setprotoent(
					    void
)
{
    enum nss_status rc;

    do_lock();
    rc = p_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&p_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getprotoent_r(
					      struct protoent *result,
					      char *buffer,
					      size_t buflen,
					      int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = p_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_PROTOCOLS, &p_state, p_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}

static enum nss_status p_read(
				 void *ni,
				 ni_idlist * match,
				 void *unused,
				 ni_proplist * quickprops,
				 void *result,
				 char *buffer,
				 size_t buflen,
				 int *errnop
)
{
    struct protoent *p = (struct protoent *) result;
    enum nss_status rc;
    ni_proplist pl;
    ni_index i;
    ni_namelist nl;
    ni_id id;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    return (NSS_STATUS_TRYAGAIN);
	}
    } else {
	pl = *quickprops;
    }
    p->p_name = (char *) STR_NULL;
    p->p_aliases = (char **) STRLST_NULL;
    p->p_proto = -1;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &p->p_name, &buffer, &buflen, errnop));
	    nl = pl.nipl_val[i].nip_val;
	    if (nl.ninl_len > 0) {
		nl.ninl_len--;
		nl.ninl_val++;
		CHECK_RC(quickprops, pl, rc, namelist_get(&nl, &p->p_aliases, &buffer, &buflen, errnop));
	    }
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_NUMBER)) {
	    int_get(&pl.nipl_val[i].nip_val, &p->p_proto);
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return NSS_STATUS_SUCCESS;
}

/*
 * NETWORKS_STUFF
 */
static const ni_name NAME_NETWORKS = "networks";
static const ni_name NAME_ADDRESS = "address";

static enum nss_status n_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);
#define n_init() do_init(NAME_NETWORKS, &n_state)

static ni_local_state n_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_getnetbyaddr_r(
					       u_long addr,
					       int type,
					       struct netent *result,
					       char *buffer,
					       size_t buflen,
					       int *errnop,
					       int *h_errnop
)
{
    enum nss_status rc;
    ni_name nm;

    do_lock();
    rc = n_init();
    if (rc == NSS_STATUS_SUCCESS) {
	struct in_addr in;
	if (type != AF_INET) {
	    return (NSS_STATUS_NOTFOUND);
	}
	in = inet_makeaddr(addr, 0);
	nm = inet_ntoa(in);
	rc = do_get(NAME_NETWORKS, &n_state, NAME_ADDRESS,
		   nm, n_read, h_errnop, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_getnetbyname_r(
					       const char *name,
					       struct netent *result,
					       char *buffer,
					       size_t buflen,
					       int *errnop,
					       int *h_errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = n_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_get(NAME_NETWORKS, &n_state, NAME_NAME,
		 name, n_read, h_errnop, result, buffer, buflen, errnop);
    }
    do_unlock();

    return rc;
}

enum nss_status _nss_netinfo_endnetent(
					  void
)
{
    enum nss_status rc;

    do_lock();
    rc = n_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&n_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setnetent(
					  void
)
{
    enum nss_status rc;

    do_lock();
    rc = n_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&n_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getnetent_r(
					    struct netent *result,
					    char *buffer,
					    size_t buflen,
					    int *errnop,
					    int *h_errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = n_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_NETWORKS, &n_state, n_read, h_errnop, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}

static enum nss_status n_read(
				 void *ni,
				 ni_idlist * match,
				 void *unused,
				 ni_proplist * quickprops,
				 void *result,
				 char *buffer,
				 size_t buflen,
				 int *errnop
)
{
    struct netent *n = (struct netent *) result;
    int *h_errnop = (int *) unused;
    enum nss_status rc;
    ni_proplist pl;
    ni_index i;
    ni_namelist nl;
    ni_id id;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    return (NSS_STATUS_TRYAGAIN);
	}
    } else {
	pl = *quickprops;
    }
    n->n_name = (char *) STR_NULL;
    n->n_aliases = (char **) STRLST_NULL;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NAME)) {
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &n->n_name, &buffer, &buflen, errnop));
	    nl = pl.nipl_val[i].nip_val;
	    if (nl.ninl_len > 0) {
		nl.ninl_len--;
		nl.ninl_val++;
		CHECK_RC(quickprops, pl, rc, namelist_get(&nl, &n->n_aliases, &buffer, &buflen, errnop));
	    }
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_ADDRESS)) {
	    nl = pl.nipl_val[i].nip_val;
	    if (nl.ninl_len > 0) {
		n->n_net = inet_network(nl.ninl_val[0]);
	    }
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return NSS_STATUS_SUCCESS;
}

/*
 * ALIAS_STUFF
 * Funny world. It was me that suggested to Ulrich Drepper that this NeXT-derived
 * API be included in glibc. (lukeh)
 */
static const ni_name NAME_ALIASES = "aliases";
static const ni_name NAME_MEMBERS = "members";

static enum nss_status alias_read(void *, ni_idlist *, void *, ni_proplist *, void *, char *, size_t, int *);
#define alias_init() do_init(NAME_ALIASES, &alias_state)

static ni_local_state alias_state =
{NI_INDEX_NULL};

enum nss_status _nss_netinfo_getaliasbyname_r(
						 const char *name,
						 struct aliasent *result,
						 char *buffer,
						 size_t buflen,
						 int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = alias_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_get(NAME_ALIASES, &alias_state, NAME_NAME,
		 name, alias_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_endaliasent(
					    void
)
{
    enum nss_status rc;

    do_lock();
    rc = alias_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_end(&alias_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_setaliasent(
					    void
)
{
    enum nss_status rc;

    do_lock();
    rc = alias_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_set(&alias_state);
    }
    do_unlock();
    return rc;
}

enum nss_status _nss_netinfo_getaliasent_r(
					      struct aliasent *result,
					      char *buffer,
					      size_t buflen,
					      int *errnop
)
{
    enum nss_status rc;

    do_lock();
    rc = alias_init();
    if (rc == NSS_STATUS_SUCCESS) {
	rc = do_getent(NAME_ALIASES, &alias_state, alias_read, NULL, result, buffer, buflen, errnop);
    }
    do_unlock();
    return rc;
}


static enum nss_status alias_read(
				     void *ni,
				     ni_idlist * match,
				     void *unused,
				     ni_proplist * quickprops,
				     void *result,
				     char *buffer,
				     size_t buflen,
				     int *errnop
)
{
    struct aliasent *alias = (struct aliasent *) result;
    enum nss_status rc;
    ni_proplist pl;
    ni_index i;
    ni_namelist nl;
    ni_id id;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    *errnop = ENETUNREACH;
	    return (NSS_STATUS_TRYAGAIN);
	}
    } else {
	pl = *quickprops;
    }
    alias->alias_name = (char *) STR_NULL;
    alias->alias_members_len = 0;
    alias->alias_members = (char **) STRLST_NULL;
    alias->alias_local = 0;
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NAME)) {
	    alias->alias_local = (ni == nichain.val[0].ni);
	    CHECK_RC(quickprops, pl, rc, name_get(&pl.nipl_val[i].nip_val, &alias->alias_name, &buffer, &buflen, errnop));
	} else if (ni_name_match(pl.nipl_val[i].nip_name,
				 NAME_MEMBERS)) {
	    alias->alias_members_len = nl.ninl_len;
	    CHECK_RC(quickprops, pl, rc, namelist_get(&pl.nipl_val[i].nip_val, &alias->alias_members, &buffer, &buflen, errnop));
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return NSS_STATUS_SUCCESS;
}

#ifdef __APPLE__
/*
 * NETGROUP_STUFF
 */
#define netgroup_init() do_init(NAME_MACHINES, &ng_state)
static const char NAME_NETGROUPS[] = "netgroups";
static ni_local_state ng_state =
{NI_INDEX_NULL};
static void *ng_read(void *, ni_idlist *, void *, ni_proplist *);

void _ni_setnetgrent(
			char *group
)
{
}

int _ni_getnetgrent(
		       char **machinep,
		       char **userp,
		       char **domainp
)
{
    return (0);
}

void _ni_endnetgrent(void)
{
}

int _ni_innetgr(
		   char *group,
		   char *host,
		   char *user,
		   char *domain
)
{
    if (!netgroup_init()) {
	return (0);
    }
    if (host == NULL) {
	return (0);
    }
    if (do_get(NAME_MACHINES, &ng_state, NAME_NAME,
	       host, ng_read, group) != NULL) {
	return (1);
    } else {
	return (0);
    }
}


static void *
 ng_read(
	    void *ni,
	    ni_idlist * match,
	    void *arg,
	    ni_proplist * quickprops
)
{
    ni_proplist pl;
    ni_index i;
    ni_id id;
    char *group = (char *) arg;

    if (quickprops == NULL) {
	id.nii_object = match->niil_val[0];
	if (ni_read(ni, &id, &pl) != NI_OK) {
	    return (NULL);
	}
    } else {
	pl = *quickprops;
    }
    for (i = 0; i < pl.nipl_len; i++) {
	if (ni_name_match(pl.nipl_val[i].nip_name, NAME_NETGROUPS)) {
	    if (ni_namelist_match(pl.nipl_val[i].nip_val,
				  group) != NI_INDEX_NULL) {
		if (quickprops == NULL) {
		    ni_proplist_free(&pl);
		}
		return (arg);
	    }
	}
    }
    if (quickprops == NULL) {
	ni_proplist_free(&pl);
    }
    return (NULL);
}

/* GRS 2/1/92 - Support for lookup_securityopt */

#define ALL		"all"
#define SECURE_PROP	"security_options"

static char domainHasFeature(void *domain, ni_id * root, const char *name,
			     char allOK);

static const char *undocumentedFeatures[] =
{
    "discourage_public_servers", NULL
};

int isSecurityOptionEnabled(const char *name)
{
    void *thisDomain;
    void *nextDomain;
    int count;
    char allOK = 1;
    ni_id root;

    for (count = 0; undocumentedFeatures[count]; count++) {
	if (!strcmp(name, undocumentedFeatures[count])) {
	    allOK = 0;
	}
    }
    if (!_ni_running() || (nichain.len < 1)) {
	return (-1);
    }
    for (count = 0; count < nichain.len; count++) {
	if (domainHasFeature(nichain.val[count].ni,
			     &nichain.val[count].root_id, name, allOK)) {
	    return 1;
	}
    }
    if (fully_expanded) {
	return 0;
    }
    /* 
       ** We've gone through all domains in the cache, but since the cache
       ** isn't necessarily complete, try to continue upwards from here.
     */
    if (ni_open(nichain.val[nichain.len - 1].ni, "..", &thisDomain) == NI_OK) {
	while (thisDomain) {
	    (void) ni_root(thisDomain, &root);
	    if (domainHasFeature(thisDomain, &root, name, allOK)) {
		ni_free(thisDomain);;
		return 1;
	    }
	    if (ni_open(thisDomain, "..", &nextDomain) != NI_OK) {
		nextDomain = NULL;
	    }
	    ni_free(thisDomain);
	    thisDomain = nextDomain;
	}
    } else {
	fully_expanded = 1;
    }
    return 0;
}

static char domainHasFeature(void *domain, ni_id * root, const char *name, char allOK)
{
    ni_namelist options;
    unsigned count;

    if (ni_lookupprop(domain, root, SECURE_PROP, &options) == NI_OK) {
	for (count = 0; count < options.ni_namelist_len; count++) {
	    if ((allOK && !strcmp(options.ni_namelist_val[count], ALL))
		|| !strcmp(options.ni_namelist_val[count], name)) {
		ni_namelist_free(&options);
		return 1;
	    }
	}
	ni_namelist_free(&options);
    }
    return 0;
}

/*
   ** GRS 2/1/92 - End of support for lookup_securityopt, start of  
   ** stuff that goes with lookup_netwareenbl
 */

#define NETWARE_DIR	"/locations/NetWare"
#define NETWARE_PROP	"enabled"
#define YES_VALUE	"YES"
#define NO_VALUE	"NO"

typedef enum _NetwareStatus {
    No, Yes, Dunno
} NetwareStatus;

static NetwareStatus domainNetwareStatus(void *domain);

int isNetwareEnabled()
{
    void *thisDomain;
    void *nextDomain;
    int count;
    NetwareStatus status;

    if (!_ni_running() || (nichain.len < 1)) {
	return (-1);
    }
    for (count = 0; count < nichain.len; count++) {
	if ((status = domainNetwareStatus(nichain.val[count].ni)) != Dunno) {
	    return (status == Yes) ? 1 : 0;
	}
    }
    /* 
       ** We've gone through all domains in the cache, but since the cache
       ** isn't necessarily complete, try to continue upwards from here.
     */
    if (ni_open(nichain.val[nichain.len - 1].ni, "..", &thisDomain) == NI_OK) {
	while (thisDomain) {
	    if ((status = domainNetwareStatus(nichain.val[count].ni))
		!= Dunno) {
		ni_free(thisDomain);
		return (status == Yes) ? 1 : 0;
	    }
	    if (ni_open(thisDomain, "..", &nextDomain) != NI_OK) {
		nextDomain = NULL;
	    }
	    ni_free(thisDomain);
	    thisDomain = nextDomain;
	}
    }
    return 0;
}

static NetwareStatus
 domainNetwareStatus(void *domain)
{
    ni_id dir;
    ni_namelist options;
    unsigned count;
    NetwareStatus ret = Dunno;

    if ((ni_root(domain, &dir) == NI_OK) &&
	(ni_pathsearch(domain, &dir, NETWARE_DIR) == NI_OK)) {
	if (ni_lookupprop(domain, &dir, NETWARE_PROP, &options) == NI_OK) {
	    for (count = 0; count < options.ni_namelist_len; count++) {
		if (!strcasecmp(options.ni_namelist_val[count], YES_VALUE)) {
		    ret = Yes;
		} else if (!strcasecmp(options.ni_namelist_val[count],
				       NO_VALUE)) {
		    ret = No;
		}
	    }
	    ni_namelist_free(&options);
	}
    }
    return ret;
}
#endif				/* __APPLE__ */
